from dataclasses import dataclass, field
import aiohttp
import asyncio
import datetime as dt

from pprint import pprint

import Library.DomoClasses.DomoAuth as dmda
import Library.DomoClasses.DomoLineage as dmdl
from Library.utils.DictDot import DictDot


from Library.DomoClasses.routes import sandbox_routes


class InvalidRepositoryError(Exception):
    pass


@dataclass
class DomoRepository:
    id: str
    name: str
    last_updated_dt: dt.datetime
    commit_dt: dt.datetime
    commit_version: str
    full_auth: dmda.DomoAuth = None
    content_page_id_ls: list[str] = None
    content_card_id_ls: list[str] = None
    content_dataflow_id_ls: list[str] = None
    content_view_id_ls: list[str] = None

    def __post_init__(self):
        self.lineage = dmdl.DomoLineage(id=self.id,
                                        parent=self)

    @classmethod
    def _from_json(cls, obj, full_auth=None):

        import dateutil.parser as dtut

        dd = DictDot(obj)

        return cls(
            id=dd.id,
            full_auth=full_auth,
            name=dd.name,
            content_page_id_ls=dd.repositoryContent.pageIds,
            content_card_id_ls=dd.repositoryContent.cardIds,
            content_dataflow_id_ls=dd.repositoryContent.dataflowIds,
            content_view_id_ls=dd.repositoryContent.viewIds,
            last_updated_dt=dtut.parse(dd.updated).replace(tzinfo=None),
            commit_dt=dtut.parse(dd.lastCommit.completed).replace(tzinfo=None),
            commit_version=dd.lastCommit.commitName
        )

    @classmethod
    async def get_from_id(cls, repository_id: str, full_auth: dmda.DomoFullAuth):
        res = await sandbox_routes.get_repo_from_id(repository_id=repository_id, full_auth=full_auth)

        if res.status == 404:
            raise InvalidRepositoryError

        return cls._from_json(res.response, full_auth=full_auth)

    def convert_lineage_to_dataframe(self, return_raw: bool = False):
        import pandas as pd
        import re

        flat_lineage_ls = self.lineage._flatten_lineage()

        output_ls = [{'sandbox_id': self.id,
                      'sandbox_name': self.name,
                      'version': self.commit_version,
                      'commit_dt': self.commit_dt,
                      'last_updated_dt': self.last_updated_dt,
                      'entity_type': row.get('entity_type'),
                      'entity_id': row.get('entity_id')
                      } for row in flat_lineage_ls]

        if return_raw:
            return output_ls

        return pd.DataFrame(output_ls)


@dataclass
class DomoSandbox:

    @classmethod
    async def get_repositories(cls, full_auth):
        res = await sandbox_routes.get_shared_repos(full_auth)

        if res.status != 200:
            return None

        domo_repos = [DomoRepository._from_json(
            obj, full_auth=full_auth) for obj in res.response.get('repositories')]

        return domo_repos



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/integrations/Automation.ipynb.

# %% auto 0
__all__ = ['remove_partition_by_x_days']

# %% ../../nbs/integrations/Automation.ipynb 2
import csv
import datetime as DT
import os
from dataclasses import dataclass

import domolibrary.client.DomoAuth as dmda
import domolibrary.classes.DomoDataset as dmds
import domolibrary.classes.DomoInstanceConfig as dmic

import utils.Exceptions as ex

import pandas as pd


# %% ../../nbs/integrations/Automation.ipynb 3
async def remove_partition_by_x_days(auth: dmda.DomoFullAuth,
                                     dataset_id: str,
                                     x_last_days: int = 0,
                                     separator: str = None,
                                     date_index: int = 0,
                                     date_format: str = '%Y-%m-%d'):
    domo_ds = dmds.DomoDataset(auth=auth, id=dataset_id)

    list_partition = await domo_ds.list_partitions(auth=auth, dataset_id=dataset_id)

    today = DT.date.today()
    days_ago = today - DT.timedelta(days=x_last_days)
    for i in list_partition:
        compare_date = ''
        if separator is not None and separator != '':
            compare_date = i['partitionId'].split(separator)[date_index]
        else:
            compare_date = i['partitionId']

        try:
            d = DT.datetime.strptime(compare_date, date_format).date()
        except ValueError:
            d = None
        if d is not None and d < days_ago:
            print(auth.domo_instance, ': 🚀  Removing partition key : ',
                  (i['partitionId']), ' in ', dataset_id)
            await domo_ds.delete_partition(dataset_partition_id=i['partitionId'], dataset_id=dataset_id,
                                           auth=auth)




import aiohttp
import pandas as pd
import io
from pprint import pprint

from .get_data import get_data, looper
from ..DomoAuth import DomoDeveloperAuth, DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


async def get_dataset_by_id(id: str,
                            full_auth: DomoFullAuth = None,
                            dev_auth: DomoDeveloperAuth = None,
                            debug: bool = False, log_result: bool = False) -> ResponseGetData:
    domo_instance = full_auth.domo_instance if full_auth else dev_auth.domo_instance

    url = f'https://{domo_instance}.domo.com/api/data/v3/datasources/{id}'

    res = await get_data(
        auth=full_auth or dev_auth,
        url=url,
        method='GET',
        debug=debug,
    )

    if log_result:
        print(res)

    return res


async def query_dataset_public(dev_auth: DomoDeveloperAuth,
                               id: str,
                               sql: str,
                               session: aiohttp.ClientSession,
                               debug: bool = False):
    url = f'https://api.domo.com/v1/datasets/query/execute/{id}?IncludeHeaders=true'

    body = {'sql': sql}

    res = await get_data(auth=dev_auth, url=url, method='POST', body=body, session=session, debug=debug)

    return res


async def query_dataset_private(full_auth: DomoFullAuth,
                                id: str,
                                sql: str,
                                session: aiohttp.ClientSession,
                                loop_until_end: bool = True,
                                limit=100,

                                debug: bool = False):
    is_close_session = False
    if not session:
        session = aiohttp.ClientSession()
        is_close_session = True

    url = f'https://{full_auth.domo_instance}.domo.com/api/query/v1/execute/{id}'
    if debug:
        print(url)

    offset_params = {
        'offset': 'offset',
        'limit': 'limit',
    }

    def body_fn(skip, limit):
        return {
            'sql': f"{sql} offset {skip} limit {limit}"
        }

    def arr_fn(res) -> pd.DataFrame:
        rows_ls = res.response.get('rows')
        columns_ls = res.response.get('columns')
        output = []
        for row in rows_ls:
            new_row = {}
            for index, column in enumerate(columns_ls):
                new_row[column] = row[index]
            output.append(new_row)
            # pd.DataFrame(data=res.response.get('rows'), columns=res.response.get('columns'))
        return output

    # res = await get_data(auth=full_auth, url=url, method='POST', body=body, session = session ,debug=debug)

    res = await looper(auth=full_auth,
                       method='POST',
                       url=url,
                       arr_fn=arr_fn,
                       offset_params=offset_params,
                       limit=limit,
                       session=session,
                       body_fn=body_fn,
                       debug=debug)

    if is_close_session:
        await session.close()

    return res


async def get_schema(full_auth: DomoFullAuth, id: str,
                     debug: bool = False, log_result: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/query/v1/datasources/{id}/schema/indexed?includeHidden=false'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        debug=debug,
    )

    if log_result:
        print(res)

    return res


# UPLOAD DATASET
async def upload_dataset_stage_1(full_auth: DomoFullAuth,
                                 dataset_id: str,
                                 session: aiohttp.ClientSession = None,
                                 debug: bool = False,
                                 restate_data_tag: str = None,
                                 data_tag: str = None
                                 ):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}/uploads"

    body = {
        "action": None,
        "appendId": None
    }
    if not restate_data_tag and not data_tag:
        params = None
    else:
        params = {'dataTag': restate_data_tag or data_tag}
        body.update({'appendId': 'latest'})

    return await get_data(auth=full_auth,
                          url=url, method='POST',
                          body=body,
                          session=session,
                          debug=debug,
                          params=params)


async def upload_dataset_stage_2_file(full_auth: DomoFullAuth,
                                      dataset_id: str,
                                      upload_id: str,
                                      file: io.TextIOWrapper = None,
                                      session: aiohttp.ClientSession = None,
                                      part_id: str = 2,
                                      debug: bool = False

                                      ):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}/uploads/{upload_id}/parts/{part_id}"
    body = file

    if debug:
        print(body)

    res = await get_data(
        url=url,
        method="PUT",
        auth=full_auth,
        content_type='text/csv',
        body=body,
        session=session,
        debug=debug
    )
    res.upload_id = upload_id
    res.dataset_id = dataset_id
    res.part_id = part_id

    return res


async def upload_dataset_stage_2_df(full_auth: DomoFullAuth,
                                    dataset_id: str,
                                    upload_id: str,
                                    upload_df: pd.DataFrame,
                                    session: aiohttp.ClientSession = None,
                                    part_id: str = 2,
                                    debug: bool = False

                                    ):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}/uploads/{upload_id}/parts/{part_id}"
    body = upload_df.to_csv(header=False, index=False)

    # if debug:
    #     print(body)

    res = await get_data(
        url=url,
        method="PUT",
        auth=full_auth,
        content_type='text/csv',
        body=body,
        session=session,
        debug=debug
    )

    res.upload_id = upload_id
    res.dataset_id = dataset_id
    res.part_id = part_id

    return res


async def upload_dataset_stage_3(full_auth: DomoFullAuth,
                                 dataset_id: str,
                                 upload_id: str,
                                 session: aiohttp.ClientSession = None,
                                 update_method: str = 'REPLACE',
                                 restate_data_tag: str = None,
                                 data_tag: str = None,
                                 is_index: bool = False,
                                 debug: bool = False):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}/uploads/{upload_id}/commit"

    body = {"index": is_index,
            "action": update_method}

    if restate_data_tag or data_tag:
        if debug:
            print('route_stage_3_updating body')

        body.update({"action": 'APPEND',
                     'dataTag': restate_data_tag or data_tag,
                     'appendId': 'latest' if (restate_data_tag or data_tag) else None,
                     'index': is_index
                     })

    res = await get_data(
        auth=full_auth,
        method="PUT",
        url=url,
        body=body,
        session=session,
        debug=debug
    )

    res.upload_id = upload_id
    res.dataset_id = dataset_id

    return res


async def upload_dataset_stage_4(full_auth: DomoFullAuth,
                                 dataset_id: str,
                                 index_id: str,
                                 session: aiohttp.ClientSession = None,
                                 debug: bool = False):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}/indexes/{index_id}/statuses"

    return await get_data(
        auth=full_auth,
        method="GET",
        url=url,
        body=body,
        session=session,
        debug=debug
    )


async def index_dataset(full_auth: DomoFullAuth,
                        dataset_id: str,
                        session: aiohttp.ClientSession = None,
                        debug: bool = False):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}/indexes"

    body = {"dataIds": []}

    return await get_data(auth=full_auth, method='POST', body=body, url=url, session=session)


def generate_list_partitions_body(limit=100, offset=0):
    return {
        "paginationFields": [{
            "fieldName": "datecompleted",
            "sortOrder": "DESC",
            "filterValues": {
                "MIN": None,
                "MAX": None
            }
        }],
        "limit": 1000,
        "offset": 0
    }


async def list_partitions(full_auth: DomoFullAuth,
                          dataset_id: str,
                          body: dict = None,
                          maximum: int = None,
                          loop_until_end: bool = True,
                          session: aiohttp.ClientSession = None,
                          debug: bool = False):
    try:

        is_close_session = False if session else True

        if not session:
            session = aiohttp.ClientSession()

        body = body or generate_list_partitions_body()

        url = f"https://{full_auth.domo_instance}.domo.com/api/query/v1/datasources/{dataset_id}/partition/list"

        offset_params = {
            'offset': 'offset',
            'limit': 'limit',
        }

        def arr_fn(res) -> list[dict]:
            return res.response

        res = await looper(auth=full_auth,
                           method='POST',
                           url=url,
                           arr_fn=arr_fn,
                           body=body,
                           offset_params_in_body=True,
                           offset_params=offset_params,
                           loop_until_end=True,
                           session=session,
                           debug=debug)

        if isinstance(res, list):
            return ResponseGetData(status=200,
                                   response=res,
                                   is_success=True)
        else:
            return ResponseGetData(status=400,
                                   response=None,
                                   is_success=False)

    finally:
        if is_close_session:
            await session.close()

# Delete partition has 3 stages
# Stage 1. This marks the data version associated with the partition tag as deleted.  It does not delete the partition tag or remove the association between the partition tag and data version.  There should be no need to upload an empty file – step #3 will remove the data from Adrenaline.


async def delete_partition_stage_1(full_auth: DomoFullAuth,
                                   dataset_id: str,
                                   dataset_partition_id: str,
                                   session: aiohttp.ClientSession = None,
                                   debug: bool = False):

    #url = f'https://{full_auth.domo_instance}.domo.com/api/query/v1/datasources/{dataset_id}/partition/{dataset_partition_id}'
    # update on 9/9/2022 based on the conversation with Greg Swensen
    url = f'https://{full_auth.domo_instance}.domo.com/api/query/v1/datasources/{dataset_id}/tag/{dataset_partition_id}/data'

    return await get_data(
        auth=full_auth,
        method="DELETE",
        url=url,
        session=session,
        debug=debug
    )
# Stage 2. This will remove the partition association so that it doesn’t show up in the list call.  Technically, this is not required as a partition against a deleted data version will not count against the 400 partition limit, but as the current partitions api doesn’t make that clear, cleaning these up will make it much easier for you to manage.


async def delete_partition_stage_2(full_auth: DomoFullAuth,
                                   dataset_id: str,
                                   dataset_partition_id: str,
                                   session: aiohttp.ClientSession = None,
                                   debug: bool = False):

    url = f'https://{full_auth.domo_instance}.domo.com/api/query/v1/datasources/{dataset_id}/partition/{dataset_partition_id}'

    return await get_data(
        auth=full_auth,
        method="DELETE",
        url=url,
        session=session,
        debug=debug
    )


async def delete(full_auth: DomoFullAuth,
                 dataset_id: str, session: aiohttp.ClientSession = None, debug: bool = False):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v3/datasources/{dataset_id}?deleteMethod=hard"

    return await get_data(
        auth=full_auth,
        method="DELETE",
        url=url,
        session=session,
        debug=debug
    )



from ..DomoClasses.DomoAuth import DomoDeveloperAuth, DomoFullAuth
import aiohttp
import asyncio
import Library.DomoClasses.DomoDataset as dmds
import pandas as pd
from datetime import datetime
import traceback


class MyLogger:
    app_name : str = ''
    output_ds : str = ''
    instance_auth : DomoFullAuth = None
    logs_df : pd.DataFrame = pd.DataFrame()
    breadcrumbs : list = []
    
    def __init__(self, app_name, output_ds, instance_auth):
        self.app_name = app_name
        self.output_ds = output_ds
        self.instance_auth = instance_auth
        breadcrumbs = []
        logs_df = pd.DataFrame()
        
    def add_crumb(self, crumb):
        if crumb not in self.breadcrumbs:
            self.breadcrumbs.append(crumb)
        
    def remove_crumb(self, crumb):
        if crumb in self.breadcrumbs:
            self.breadcrumbs.remove(crumb)

    def log_info(self, message, is_print = True):
        self.__AddLog(message = message,
                      type_str = "Info", 
                     is_print=is_print)

    def log_error(self, message, is_print = True):
        self.__AddLog(message = message,
                      type_str = "Error", 
                     is_print=is_print)

    def log_warning(self, message, is_print = True):
        self.__AddLog(message = message,
                      type_str = "Warning", 
                     is_print=is_print)


    def __AddLog(self, message: str, type_str : str, is_print = True):
        traceback_ls = []
        
        for line in traceback.format_stack():
            if not line.strip().startswith('File "/home/domo/.conda'):
                traceback_ls.append(line.strip())
        
        new_row = pd.DataFrame({'date_time':datetime.now(), 'application':self.app_name, 'type':type_str, 'message': message, 'breadcrumbs': None if self.breadcrumbs is None else '->'.join(self.breadcrumbs), 'traceback':None if traceback_ls is None else ' '.join(traceback_ls)}, index=[0])
        if is_print:
            print (message)
        self.logs_df = pd.concat([new_row,self.logs_df.loc[:]]).reset_index(drop=True)
    
    async def write_logs (self, upload_method: str = 'APPEND'):
        dataset = dmds.DomoDataset(full_auth = self.instance_auth,
                                    id = self.output_ds)
        await dataset.upload_csv(upload_df = self.logs_df,upload_method = upload_method)
        await asyncio.sleep(10)
        await dataset.index_dataset()
        #remove all rows
        self.logs_df = self.logs_df.head(0)
        print ('sucess, logs are saved')




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoApplication.ipynb.

# %% auto 0
__all__ = []

# %% ../../nbs/classes/50_DomoApplication.ipynb 2
import domolibrary.routes.job as job_routes
import domolibrary.routes.application as application_routes
import domolibrary.classes.DomoJob as dmdj
import pandas as pd
from dataclasses import dataclass

from pprint import pprint

import datetime as dt
from dataclasses import dataclass, field
from typing import List, Optional
from enum import Enum

import json
import io

import httpx
import asyncio

import domolibrary.utils.DictDot as DictDot
import domolibrary.client.DomoAuth as dmda

from enum import Enum



# %% ../../nbs/classes/50_DomoApplication.ipynb 3
@dataclass
class DomoApplication:
    id: str
    customer_id: str = None
    name: str = None
    description: str = None
    version: str = None
    execution_class: str = None
    grants: [str] = None
    jobs: [dmdj.DomoJob] = field(default=None)
    jobs_schedule: pd.DataFrame = field(default=None, repr=False)
    auth: dmda.DomoFullAuth = field(repr=False, default=None)

    @classmethod
    def _from_json(cls, obj, auth: dmda.DomoFullAuth = None):
        dd = DictDot(obj)

        return cls(
            id=dd.applicationId,
            customer_id=dd.customerId,
            name=dd.name,
            description=dd.description,
            version=dd.version,
            execution_class=dd.executionClass,
            grants=dd.authorities,
            auth=auth)

    @classmethod
    async def get_from_id(cls, auth: dmda.DomoFullAuth, application_id=None):
        res = await application_routes.get_application_by_id(application_id=application_id, auth=auth or self.auth)

        if res.status == 200:
            return cls._from_json(obj=res.response, auth=auth)

    async def get_jobs(self, auth: dmda.DomoFullAuth = None,
                       application_id: str = None,
                       debug_api: bool = False, 
                       session: Optional[httpx.AsyncClient] = None, 
                       return_raw: bool = False):

        res = await job_routes.get_jobs(auth=auth or self.auth,
                                        application_id=application_id or self.id,
                                        debug_api=debug_api,
                                        session=session)
        if debug_api:
            print('Getting Domostats jobs')

        if res.status == 200 and not return_raw:
            self.jobs = [dmdj.DomoJob._from_json(job) for job in res.response]
            return self.jobs

        if res.status == 200 and return_raw:
            return res.response

    async def get_all_schedules(self, auth: dmda.DomoFullAuth = None):
        if not self.jobs and (self.auth or auth):
            await self.get_jobs()

        elif not self.jobs and not (self.auth or auth):
            raise Exception("pass a auth object")

        schedules = pd.DataFrame([{'hour': trigger.schedule.hour,
                                   'minute': trigger.schedule.minute,
                                   'job_id': job.id,
                                   'job_name': job.name,
                                   'trigger_id': trigger.id} for job in self.jobs for trigger in job.triggers])

        self.jobs_schedule = schedules.sort_values(
            ['hour', 'minute'], ascending=True).reset_index(drop=True)
        return self.jobs_schedule

    async def find_next_job_schedule(self) -> dmdj.DomoTrigger_Schedule:
        if not isinstance(self.jobs_schedule, pd.DataFrame) and (self.auth or auth):
            await self.get_all_schedules()

        elif not isinstance(self.jobs_schedule, pd.DataFrame) and not (self.auth or auth):
            raise Exception("pass a auth object")

        df_all_hours = pd.DataFrame(range(0, 23), columns=['hour'])
        df_all_minutes = pd.DataFrame(range(0, 60), columns=['minute'])

        df_all_hours['tmp'] = 1
        df_all_minutes['tmp'] = 1
        df_all = pd.merge(df_all_hours, df_all_minutes,
                          on='tmp').drop(columns=['tmp'])

        # get the number of occurencies of each hour and minutes
        schedules_grouped = self.jobs_schedule.groupby(
            ['hour', 'minute']).size().reset_index(name='cnt_schedule')

        # print(schedules_grouped)
        # print(df_all)

        schedules_interpolated = pd.merge(
            df_all, schedules_grouped, how='left', on=['hour', 'minute'])

        schedules_interpolated['cnt_schedule'] = schedules_interpolated['cnt_schedule'].fillna(
            value=0)
        schedules_interpolated.sort_values(
            ['cnt_schedule', 'hour', 'minute'], ascending=True, inplace=True)

        schedules_interpolated.reset_index(drop=True, inplace=True)

        return dmdj.DomoTrigger_Schedule(
            hour=int(schedules_interpolated.loc[0].get('hour')),
            minute=int(schedules_interpolated.loc[0].get('minute')))



from .get_data import get_data
from ..DomoAuth import DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


async def create_user(full_auth: DomoFullAuth, display_name, email, role_id, debug: bool = False,
                      log_results: bool = False):
    url = f"https://{full_auth.domo_instance}.domo.com/api/content/v3/users"

    body = {"displayName": display_name, "detail": {
        "email": email}, "roleId": role_id}

    res = await get_data(url=url, method='POST', auth=full_auth, body=body, log_results=log_results, debug=debug)
    return res


async def get_all_users(full_auth: DomoFullAuth, log_results: bool = False, debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/content/v2/users'

    if log_results:
        print(f'Getting users from - {url}')

    res = await get_data(url=url, method='GET', auth=full_auth, log_results=log_results, debug=debug)

    return res


def generate_search_users_body_by_id(ids: list[str]) -> dict:
    return {
        # "showCount": true,
        # "count": false,
        "includeDeleted": False,
        "includeSupport": False,
        "filters": [
            {
                "field": "id",
                "filterType": "value",
                "values": ids,
                "operator": "EQ"

            }
        ]
    }


def generate_search_users_body_by_email(email_address) -> dict:
    """search does not appear to be case sensitive"""

    body = {
        # "showCount": true,
        # "count": false,
        "includeDeleted": False,
        "includeSupport": False,
        "limit": 200,
        "offset": 0,
        "sort": {
            "field": "displayName",
            "order": "ASC"
        },
        "filters": [
            {
                "filterType": "text",
                "field": "emailAddress",
                "text": email_address
            }
        ]

    }
    return body


async def search_virtual_user_by_subscriber_instance(full_auth: DomoFullAuth, subscriber_instance, debug: False,
                                                     log_results: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/publish/v2/proxy_user/domain/'

    body = {"domains": [f"{subscriber_instance}.domo.com"]}

    res = await get_data(url=url, method='POST', auth=full_auth, body=body, log_results=log_results, debug=debug)
    return res


async def search_users(full_auth: DomoFullAuth, body: dict, log_results: bool = False,
                       debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/identity/v1/users/search'

    if log_results:
        print(f'Getting users from - {url}')

    res = await get_data(url=url, method='POST', auth=full_auth, body=body, log_results=log_results, debug=debug)

    return res


async def reset_password(full_auth: DomoFullAuth, user_id: str, new_password: str, log_results: bool = False,
                         debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/identity/v1/password'

    body = {
        "domoUserId": user_id,
        "password": new_password}

    if debug:
        print(url)
        print(body)

    res = await get_data(url=url, method='PUT', auth=full_auth, body=body, log_results=log_results, debug=debug)

    return res


async def request_password_reset(domo_instance: str, email: str, locale='en-us', debug: bool = False):
    url = f'https://{domo_instance}.domo.com/api/domoweb/auth/sendReset'

    params = {'email': email,
              'local': locale}

    return await get_data(url=url,
                          method='GET',
                          auth=None,
                          params=params, debug=debug)


async def set_user_landing_page( full_auth: DomoFullAuth, user_id: str,
                             page_id:str,
                             debug:bool = False ):
    
    url = f'https://{full_auth.domo_instance}.domo.com/api/content/v1/landings/target/DESKTOP/entity/PAGE/id/{page_id}/{user_id}'
        
    return await get_data(url=url,
                          method='PUT',
                          auth=full_auth,
                          # body = body,
                          debug=debug)

                             
                             


# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/client/95_Logger.ipynb.

# %% auto 0
__all__ = ['Logger', 'TracebackDetails']

# %% ../../nbs/client/95_Logger.ipynb 3
import datetime as dt

from typing import Optional, List
from dataclasses import dataclass

import traceback

from fastcore.basics import patch_to

# %% ../../nbs/client/95_Logger.ipynb 4
class Logger:
    """log class with user customizeable output method"""

    root_module: str
    app_name: str

    logs: List[dict]
    breadcrumb: Optional[list]

    entity_id: Optional[str]
    domo_instance: Optional[str]
    # function to call with write_logs method.
    output_fn: Optional[callable] = None

    def __init__(
        self,
        app_name: str,  # name of the app for grouping logs
        root_module: Optional[str] = "<module>",  # root module for stack trace
        output_fn: Optional[
            callable
        ] = None,  # function to call with write_logs method.
        entity_id: Optional[str] = None,
        domo_instance: Optional[str] = None,
    ):

        self.app_name = app_name
        self.output_fn = output_fn
        self.root_module = root_module
        self.logs = []
        self.breadcrumb = []
        self.domo_instance = domo_instance
        self.entity_id = entity_id

    def _add_crumb(self, crumb):
        if crumb not in self.breadcrumb:
            self.breadcrumb.append(crumb)

    def _remove_crumb(self, crumb):
        if crumb in self.breadcrumb:
            self.breadcrumb.remove(crumb)

# %% ../../nbs/client/95_Logger.ipynb 6
@patch_to(Logger)
def _get_traceback(
    self :Logger,
    root_module: str = "<module>",
    num_stacks_to_drop=0,  # drop entries from the top of stack to exclude the functions that retrieve the traceback
) -> [
    traceback.FrameSummary
]:  # returns a filtered list of FrameSummaries from traceback
    """method that retrieves traceback"""

    tb = traceback.extract_stack()

    # find the last module index
    module_index = 0
    for index, tb_line in enumerate(tb):
        function_name = tb_line[2]

        if function_name == root_module:
            module_index = index

    if num_stacks_to_drop == 0:
        return tb[module_index:]

    return tb[module_index:-num_stacks_to_drop]

# %% ../../nbs/client/95_Logger.ipynb 8
@dataclass
class TracebackDetails:
    """result of _get_traceback_details function"""

    function_name: str
    file_name: str
    function_trail: str


@patch_to(Logger)
def _get_traceback_details(
    self : Logger,
    traceback_list: [
        traceback.FrameSummary
    ],  # clean list of frame summaries from traceback (be sure to exclude frames from functions that retrieved the traceback)
) -> TracebackDetails:  # descriptive summary from the top of the traceback
    """returns TracebackDetails, for the entry at the top of the stack"""

    function_trail = " -> ".join([line[2] for line in traceback_list])

    function_name = traceback_list[-1][2]
    file_name = traceback_list[-1][0]

    return TracebackDetails(function_name, file_name, function_trail)

# %% ../../nbs/client/95_Logger.ipynb 14
@patch_to(Logger)
def _add_log(
    self :Logger, message: str, type_str: str, 
    debug_log : bool = False,
    num_stacks_to_drop=3, entity_id: Optional[str] = None, domo_instance: Optional[str] = None
) -> dict:
    """internal method to append message to log"""

    traceback_list = self._get_traceback(num_stacks_to_drop=num_stacks_to_drop)
    
    if debug_log:
        print({"num_stacks_to_drop": num_stacks_to_drop})
        print([tb_line[2] for tb_line in traceback_list])

    new_row = {
        "date_time": dt.datetime.now(),
        "application": self.app_name,
        "log_type": type_str,
        "log_message": message,
        "breadcrumb": "->".join(self.breadcrumb),
        "domo_instance": domo_instance or self.domo_instance,
        "entity_id": entity_id or self.entity_id
    }

    traceback_details = self._get_traceback_details(traceback_list)

    new_row.update(
        {
            "function_name": traceback_details.function_name,
            "file_name": traceback_details.file_name,
            "function_trail": traceback_details.function_trail,
        }
    )

    if debug_log:
        print(new_row)

    self.logs.append(new_row)

    return new_row


@patch_to(Logger)
def log_info(self : Logger, message,
             entity_id: Optional[str] = None,
             domo_instance: Optional[str] = None,
             debug_log=False, num_stacks_to_drop=3):
    """log an informational message"""
    return self._add_log(
        message=message,
        entity_id=entity_id,
        domo_instance=domo_instance,
        type_str="Info",
        num_stacks_to_drop=num_stacks_to_drop,
        debug_log=debug_log,
    )


@patch_to(Logger)
def log_error(self: Logger, message,
              entity_id: Optional[str] = None,
              domo_instance: Optional[str] = None, debug_log=False, num_stacks_to_drop=3):
    """log an error message"""

    return self._add_log(
        message=message,
        entity_id = entity_id,
        domo_instance = domo_instance,
        type_str="Error",
        num_stacks_to_drop=num_stacks_to_drop,
        debug_log=debug_log,
    )


@patch_to(Logger)
def log_warning(self :Logger, message,
                entity_id: Optional[str] = None,
                domo_instance: Optional[str] = None, 
                debug_log = False,
                num_stacks_to_drop=3):
    """log a warning message"""

    return self._add_log(
        message=message,
        entity_id=entity_id,
        domo_instance=domo_instance,
        type_str="Warning",
        num_stacks_to_drop=num_stacks_to_drop,
        debug_log = debug_log,
    )


# %% ../../nbs/client/95_Logger.ipynb 17
@patch_to(Logger)
def output_log(self:Logger):
    """calls the user defined output function"""
    return self.output_fn(self.logs)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/chunk_execution.ipynb.

# %% auto 0
__all__ = ['run_sequence', 'chunk_list']

# %% ../../nbs/utils/chunk_execution.ipynb 2
from typing import Any, Awaitable

# %% ../../nbs/utils/chunk_execution.ipynb 3
async def run_sequence(*functions: Awaitable[Any] # comma separated list of functions
 ) -> None: # no explicit return
    """executes a sequence of functions"""

    return [ await function for function in functions]
        

# %% ../../nbs/utils/chunk_execution.ipynb 6
def chunk_list(obj_ls :list[any],  # list of entities to split into n chunks
               chunk_size:int  # entities per sub list
               ) -> list[list[dict]]:  # returns a list of chunk_size lists of objects

    return [obj_ls[i * chunk_size:(i + 1) * chunk_size] for i in range((len(obj_ls) + chunk_size - 1) // chunk_size)]

